/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 *
 *
 * Copyright (c) 1988 by Sun Microsystems, Inc.
 */

/*
 * Command to change one's public key in the public key database
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpcsvc/yp_prot.h>
#include <rpcsvc/ypclnt.h>

#include <gmp.h>      /* GNU multiprecission library */
#undef min
#include <mp.h>

#include "../src/rpc_mac.h"


#if 0
  #include <pwd.h>
#else
  struct passwd {
         char *pw_name;
         uid_t pw_uid;
         gid_t pw_gid;
         char *pw_dir;
         char *pw_shell;
         char *pw_passwd;
       };
  extern struct passwd * _getpwuid (int id);
#endif

/*
 * Prototypes
 */
extern char *crypt        (char*, char*);
extern int   key_setsecret(char*);
static void  usage        (char *name);
static void  genkeys      (char *public, char *secret, char *pass);
static void  adjust       (char *keyout, char *keyin);
static u_int setpublicmap (char *name, char *public, char *secret);
static struct passwd *ypgetpwuid (int);

static char PKMAP[] = "publickey.byname";
static char *domain;

int main (int argc, char **argv)
{
  char name  [MAXNETNAMELEN + 1];
  char public[HEXKEYBYTES + 1];
  char secret[HEXKEYBYTES + 1];
  char crypt1[HEXKEYBYTES + KEYCHECKSUMSIZE + 1];
  char crypt2[HEXKEYBYTES + KEYCHECKSUMSIZE + 1];

  struct passwd *pw;
  char  *pass, *master, *self;
  int    status, euid, force;

  self  = argv[0];
  force = 0;
  for (argc--, argv++; argc > 0 && **argv == '-'; argc--, argv++)
  {
    if (argv[0][2] != 0)
       usage (self);

    switch (argv[0][1])
    {
      case 'f':
           force = 1;
           break;
      default:
           usage (self);
    }
  }
  if (argc)
     usage (self);

  yp_get_default_domain (&domain);
  if (yp_master(domain, PKMAP, &master))
  {
    fprintf (stderr,"can't find master of publickey database\n");
    return (1);
  }

  getnetname (name);
  printf ("Generating new key for %s.\n", name);

  euid = __geteuid();
  if (euid)
  {
    pw = ypgetpwuid (euid);
    if (pw == NULL)
    {
      fprintf (stderr,"No yp password found: can't change key.\n");
      return (1);
    }
  }
  else
  {
    pw = _getpwuid (0);
    if (pw == NULL)
    {
      fprintf (stderr, "No password found: can't change key.\n");
      return (1);
    }
  }
  pass = getpass ("Password:");
  if (!force)
  {
    if (strcmp(crypt(pass, pw->pw_passwd), pw->pw_passwd))
    {
      fprintf (stderr, "Invalid password.\n");
      return (1);
    }
  }

  genkeys (public, secret, pass);

  memcpy (crypt1,secret,HEXKEYBYTES);
  memcpy (crypt1 + HEXKEYBYTES,secret,KEYCHECKSUMSIZE);
  crypt1 [HEXKEYBYTES + KEYCHECKSUMSIZE] = 0;
  xencrypt (crypt1, pass);

  if (force)
  {
    memcpy (crypt2, crypt1, HEXKEYBYTES + KEYCHECKSUMSIZE + 1);
    xdecrypt (crypt2, getpass("Retype password:"));
    if (memcmp(crypt2, crypt2 + HEXKEYBYTES, KEYCHECKSUMSIZE) ||
        memcmp(crypt2, secret, HEXKEYBYTES))
    {
      fprintf (stderr, "Password incorrect.\n");
      return (1);
    }
  }

  printf ("Sending key change request to %s...\n", master);
  status = setpublicmap (name, public, crypt1);
  if (status)
  {
    printf ("%s: unable to update yp database (%u): %s\n",
            self, status, yperr_string(status));
    printf ("Perhaps %s is down?\n", master);
    return (1);
  }
  printf ("Done.\n");

  if (key_setsecret(secret) < 0)
  {
    printf ("Unable to login with new secret key.\n");
    return (1);
  }
  return (0);
}

void usage (char *name)
{
  fprintf (stderr, "usage: %s [-f]\n", name);
  exit (1);
}


/*
 * Generate a seed
 */
void getseed (char *seed, int seedsize, unsigned char *pass)
{
  int    i, rseed;
  struct timeval tv;

  gettimeofday (&tv,NULL);
  rseed = tv.tv_sec + tv.tv_usec;
  for (i = 0; i < 8; i++)
      rseed ^= (rseed << 8) | pass[i];

  srandom (rseed);

  for (i = 0; i < seedsize; i++)
     seed[i] = (random() & 0xff) ^ pass[i % 8];
}


/*
 * Generate a random public/secret key pair
 */
void genkeys (char *public, char *secret, char *pass)
{
  int i;
 
#define BASEBITS  (8*sizeof(short) - 1)
#define BASE      (1 << BASEBITS)

  MINT *pk   = itom (0);
  MINT *sk   = itom (0);
  MINT *base = itom(BASE);
  MINT *root = itom(PROOT);
  MINT *modulus = xtom(HEXMODULUS);
  MINT *tmp;
  short r;
  unsigned short seed[KEYSIZE/BASEBITS + 1];
  char *xkey;

  getseed ((char*)seed, sizeof(seed), (unsigned char *)pass);
  for (i = 0; i < KEYSIZE/BASEBITS + 1; i++)
  {
    r = seed[i] % BASE;
    tmp = itom (r);
#ifdef BERKELEY_MP
    mult (sk,base, sk);
    madd (sk,tmp, sk);
#else
    mpz_mul (sk,base, sk);
    mpz_add (sk,tmp, sk);
#endif
    mfree (tmp);
  }
  tmp = itom (0);
  mdiv (sk, modulus, tmp, sk);
  mfree (tmp);
  pow (root, sk, modulus, pk);
  xkey = mtox (sk);
  adjust (secret, xkey);
  xkey = mtox (pk);
  adjust (public, xkey);
  mfree (sk);
  mfree (base);
  mfree (pk);
  mfree (root);
  mfree (modulus);
} 

/*
 * Adjust the input key so that it is 0-filled on the left
 */
void adjust (char *keyout, char *keyin)
{
  char *p, *s;

  for (p = keyin; *p; p++)
      ;
  for (s = keyout + HEXKEYBYTES; p >= keyin; p--, s--)
      *s = *p;

  while (s >= keyout)
        *s-- = '0';
}

/*
 * Set the entry in the public key map
 */
u_int setpublicmap (char *name, char *public, char *secret)
{
  char  pkent[1024];
  u_int rslt;

  sprintf (pkent, "%s:%s", public, secret);
  rslt = yp_update (domain, PKMAP, YPOP_STORE,
                    name, strlen(name), pkent, strlen(pkent));
  return (rslt);
}

struct passwd *ypgetpwuid (int uid)
{
  char   uidstr[10];
  char  *val, *p;
  int    vallen;
  static struct passwd pw;

  sprintf (uidstr, "%d", uid);
  if (yp_match(domain,"passwd.byuid",uidstr,strlen(uidstr),&val,&vallen))
     return (NULL);

  p = strchr (val, ':');
  if (p == NULL)
     return (NULL);

  pw.pw_passwd = p + 1;
  p = strchr (pw.pw_passwd, ':');
  if (p == NULL)
     return (NULL);

  *p = 0;
  return (&pw);
}
