/* @(#)rpc_hout.c  2.1 88/08/01 4.0 RPCSRC */
/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * rpc_hout.c, Header file outputter for the RPC protocol compiler 
 * Copyright (C) 1987, Sun Microsystems, Inc. 
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "rpc_scan.h"
#include "rpc_pars.h"
#include "rpc_util.h"

static void pdeclaration (char *name, declaration *dec, int indent);
static void pprocdef     (proc_list *proc, version_list *vp);

static void pstructdef (definition *def)
{
  decl_list *l;
  char      *name = def->def_name;

  fprintf (fout, "typedef struct %s {\n", name);
  for (l = def->def.st.decls; l; l = l->next)
      pdeclaration (name, &l->decl, 8);

  fprintf (fout, "      } %s;\n", name);
}

static void puniondef (definition *def)
{
  case_list   *l;
  char        *name = def->def_name;
  declaration *decl;

  fprintf (fout, "typedef struct %s {\n", name);
  decl = &def->def.un.enum_decl;
  if (streq(decl->type, "bool"))
       fprintf (fout, "        bool_t %s;\n", decl->name);
  else fprintf (fout, "        %s %s;\n", decl->type, decl->name);

  fprintf (fout, "  union {\n");
  for (l = def->def.un.cases; l; l = l->next)
      pdeclaration (name, &l->case_decl, 4);

  decl = def->def.un.default_decl;
  if (decl && !streq(decl->type, "void"))
     pdeclaration (name, decl, 4);

  fprintf (fout, "    } %s_u;\n", name);
  fprintf (fout, "  } %s;\n",name);
}


static void pdefine (char *name, char *num)
{
  fprintf (fout, "#define %-25s %s\n", name, num);
}

static void puldefine (char *name, char *num)
{
  fprintf (fout, "#define %-25s %sUL\n", name, num);
}

static void pconstdef (definition *def)
{
  pdefine (def->def_name, def->def.co);
}

static int define_printed (proc_list *stop, version_list *start)
{
  version_list *vers;
  proc_list    *proc;

  for (vers = start; vers; vers = vers->next)
      for (proc = vers->procs; proc; proc = proc->next)
      {
        if (proc == stop)
           return (0);
        if (streq(proc->proc_name, stop->proc_name))
           return (1);
      }
  abort();
  /* NOTREACHED */
  return (0);
}


static int undefined2 (char *type, char *stop)
{
  list       *l;
  definition *def;

  for (l = defined; l; l = l->next)
  {
    def = (definition *) l->val;
    if (def->def_kind != DEF_PROGRAM)
    {
      if (streq(def->def_name, stop))
         return (1);
      if (streq(def->def_name, type))
         return (0);
    }
  }
  return (1);
}


static void pprogramdef (definition *def)
{
  version_list *vers;
  proc_list    *proc;

  puldefine (def->def_name, def->def.pr.prog_num);
  for (vers = def->def.pr.versions; vers; vers = vers->next)
  {
    puldefine (vers->vers_name, vers->vers_num);
    for (proc = vers->procs; proc; proc = proc->next)
    {
      if (!define_printed(proc, def->def.pr.versions))
         puldefine (proc->proc_name, proc->proc_num);
      pprocdef (proc, vers);
      fputc ('\n',fout);
    }
  }
}


static void pprocdef (proc_list *proc, version_list *vp)
{
  fprintf (fout, "extern ");
  if (proc->res_prefix)
  {
    if (streq(proc->res_prefix, "enum"))
         fprintf (fout, "enum ");
    else fprintf (fout, "struct ");
  }
  if (streq(proc->res_type, "bool"))
       fprintf (fout, "bool_t *");
  else if (streq(proc->res_type, "string"))
       fprintf (fout, "char **");
  else fprintf (fout, "%s *", fixtype(proc->res_type));

  pvname (proc->proc_name, vp->vers_num);
  fprintf (fout, "();\n");
}


static void penumdef (definition *def)
{
  enumval_list *l;
  char         *name = def->def_name;
  char         *last  = NULL;
  int           count = 0;

  fprintf (fout, "typedef enum %s {\n", name);
  for (l = def->def.en.vals; l; l = l->next)
  {
    fprintf (fout, "          %s", l->name);
    if (l->assignment)
    {
      fprintf (fout, " = %s", l->assignment);
      last  = l->assignment;
      count = 1;
    }
    else
    {
      if (last == NULL)
           fprintf (fout, " = %d", count++);
      else fprintf (fout, " = %s + %d", last, count++);
    }
    fprintf (fout, ",\n");
  }
  fprintf (fout, "      } %s;\n", name);
}


static void ptypedef (definition *def)
{
  char     prefix[8];
  char    *name = def->def_name;
  char    *old  = def->def.ty.old_type;
  relation rel  = def->def.ty.rel;

  if (!streq(name, old))
  {
    if (streq(old, "string"))
    {
      old = "char";
      rel = REL_POINTER;
    }
    else if (streq(old, "opaque"))
         old = "char";
    else if (streq(old, "bool"))
         old = "bool_t";

    if (undefined2(old, name) && def->def.ty.old_prefix)
         sprintf (prefix, "%s ", def->def.ty.old_prefix);
    else prefix[0] = 0;

    fprintf (fout, "typedef ");
    switch (rel)
    {
      case REL_ARRAY:
           fprintf (fout, "struct {\n");
           fprintf (fout, "  %s %s_len;\n", INT32_NAM,name);
           fprintf (fout, "  %s%s *%s_val;\n", prefix, old, name);
           fprintf (fout, "} %s", name);
           break;

      case REL_POINTER:
           fprintf (fout, "%s%s *%s", prefix, old, name);
           break;

      case REL_VECTOR:
           fprintf (fout, "%s%s %s[%s]", prefix, old, name,
                    def->def.ty.array_max);
           break;

      case REL_ALIAS:
           fprintf (fout, "%s%s %s", prefix, old, name);
           break;
    }
    fprintf (fout, ";\n");
  }
}


static void pdeclaration (char *name, declaration *dec, int indent)
{
  char  buf[8];
  char *prefix;
  char *type;

  if (streq(dec->type, "void"))
     return;

  fprintf (fout,"%*s", indent,"");

  if (streq(dec->type, name) && !dec->prefix)
     fprintf (fout, "struct ");

  if (streq(dec->type, "string"))
     fprintf (fout, "char *%s", dec->name);
  else
  {
    prefix = "";
    if (streq(dec->type, "bool"))
       type = "bool_t";
    else if (streq(dec->type, "opaque"))
       type = "char";
    else
    {
      if (dec->prefix)
      {
        sprintf (buf, "%s ", dec->prefix);
        prefix = buf;
      }
      type = dec->type;
    }
    switch (dec->rel)
    {
      case REL_ALIAS:
           fprintf (fout, "%s%s %s", prefix, type, dec->name);
           break;
      case REL_VECTOR:
           fprintf (fout, "%s%s %s[%s]", prefix, type, dec->name,
                    dec->array_max);
           break;
      case REL_POINTER:
           fprintf (fout, "%s%s *%s", prefix, type, dec->name);
           break;
      case REL_ARRAY:
           fprintf (fout,
                    "struct {\n"
                    "%*s  %s %s_len;\n"
                    "%*s  %s%s *%s_val;\n%*s} %s",
                    indent,"", INT32_NAM,dec->name,
                    indent,"", prefix, type, dec->name,
                    indent,"", dec->name);
           break;
    }
  }
  fprintf (fout, ";\n");
}

/*
 * Print the C-version of an xdr definition 
 */
void print_datadef (definition *def)
{
  if (def->def_kind != DEF_CONST)
     fprintf (fout, "\n");

  switch (def->def_kind)
  {
    case DEF_STRUCT:
         pstructdef (def);
         break;
    case DEF_UNION:
         puniondef (def);
         break;
    case DEF_ENUM:
         penumdef (def);
         break;
    case DEF_TYPEDEF:
         ptypedef (def);
         break;
    case DEF_PROGRAM:
         pprogramdef (def);
         break;
    case DEF_CONST:
         pconstdef (def);
         break;
  }
  if (def->def_kind != DEF_PROGRAM && def->def_kind != DEF_CONST)
     fprintf (fout, "bool_t xdr_%s();\n", def->def_name);

  if (def->def_kind != DEF_CONST)
     fprintf (fout, "\n");
}

