/* @(#)rpc_svcout.c  2.1 88/08/01 4.0 RPCSRC */
/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * rpc_svcout.c, Server-skeleton outputter for the RPC protocol compiler
 * Copyright (C) 1987, Sun Microsytsems, Inc. 
 */

#include <stdio.h>
#include <string.h>

#include "rpc_scan.h"
#include "rpc_pars.h"
#include "rpc_util.h"

static char RQSTP[]   = "rqstp";
static char TRANSP[]  = "transp";
static char ARG[]     = "argument";
static char RESULT[]  = "result";
static char ROUTINE[] = "local";


static void printerr (char *err, char *transp, int indent)
{
  fprintf (fout, "%*ssvcerr_%s (%s);\n", indent,"", err, transp);
}

static void printif (char *proc, char *transp, char *prefix, char *arg)
{
  fprintf (fout, "  if (!svc_%s(%s, xdr_%s, %s%s))\n  {\n",
           proc, transp, arg, prefix, arg);
}

static int nullproc (proc_list *proc)
{
  for (; proc; proc = proc->next)
      if (streq(proc->proc_num, "0"))
         return (1);
  return (0);
}


/*
 * write most of the service, that is, everything but the registrations. 
 */
void write_most (void)
{
  list         *l;
  definition   *def;
  version_list *vp;

  for (l = defined; l; l = l->next)
  {
    def = (definition *) l->val;
    if (def->def_kind == DEF_PROGRAM)
    {
      for (vp = def->def.pr.versions; vp; vp = vp->next)
      {
        fprintf (fout, "\nstatic void ");
        pvname (def->def_name, vp->vers_num);
        fprintf (fout, "();");
      }
    }
  }

  fprintf (fout, "\n\nint main (void)\n{\n"\
                 "  SVCXPRT *%s;\n\n", TRANSP);

  for (l = defined; l; l = l->next)
  {
    def = (definition*) l->val;
    if (def->def_kind == DEF_PROGRAM)
       for (vp = def->def.pr.versions; vp; vp = vp->next)
           fprintf (fout, "  pmap_unset (%s, %s);\n",
                    def->def_name, vp->vers_name);
  }
}


/*
 * write a registration for the given transport 
 */
void write_register (char *transp)
{
  list         *l;
  definition   *def;
  version_list *vp;

  fprintf (fout, "\n  %s = svc%s_create (RPC_ANYSOCK", TRANSP, transp);
  if (streq(transp, "tcp"))
     fprintf (fout, ", 0, 0");

  fprintf (fout, ");\n  if (%s == NULL)\n  {\n", TRANSP);
  fprintf (fout, "    fprintf (stderr, \"cannot create %s service.\\n\");\n", transp);
  fprintf (fout, "    return (1);\n");
  fprintf (fout, "  }\n");

  for (l = defined; l; l = l->next)
  {
    def = (definition*) l->val;
    if (def->def_kind != DEF_PROGRAM)
       continue;

    for (vp = def->def.pr.versions; vp; vp = vp->next)
    {
      fprintf (fout, "  if (!svc_register(%s, %s, %s, ",
               TRANSP, def->def_name, vp->vers_name);
      pvname (def->def_name, vp->vers_num);
      fprintf (fout, ", IPPROTO_%s))\n  {\n",
               streq (transp, "udp") ? "UDP" : "TCP");
      fprintf (fout,
               "    fprintf (stderr, \"unable to register (%s,%s,%s).\\n\");\n",
               def->def_name, vp->vers_name, transp);
      fprintf (fout, "    return (1);\n");
      fprintf (fout, "  }\n");
    }
  }
}

/*
 * write the rest of the service 
 */
void write_rest (void)
{
  fprintf (fout,
           "  svc_run();\n"
           "  fprintf (stderr, \"svc_run returned\\n\");\n"
           "  return (1);\n"
           "}\n");
}


static void write_program (definition *def, char *storage)
{
  version_list *vp;
  proc_list    *proc;
  int          filled;

  for (vp = def->def.pr.versions; vp; vp = vp->next)
  {
    fprintf (fout,"\n");
    if (storage)
       fprintf (fout, "%s ", storage);

    fprintf (fout, "void ");
    pvname (def->def_name, vp->vers_num);
    fprintf (fout, " (struct svc_req *%s, SVCXPRT *%s)\n{\n", RQSTP, TRANSP);

    filled = 0;
    fprintf (fout, "  union {\n");
    for (proc = vp->procs; proc; proc = proc->next)
    {
      if (streq(proc->arg_type, "void"))
         continue;

      filled = 1;
      fprintf (fout, "    ");
      ptype (proc->arg_prefix, proc->arg_type, 0);
      pvname (proc->proc_name, vp->vers_num);
      fprintf (fout, "_arg;\n");
    }
    if (!filled)
       fprintf (fout, "    int fill;\n");

    fprintf (fout, "  } %s;\n\n", ARG);
    fprintf (fout, "  char *%s;\n", RESULT);
    fprintf (fout, "  bool_t (*xdr_%s)(), (*xdr_%s)();\n", ARG, RESULT);
    fprintf (fout, "  char *(*%s)();\n\n", ROUTINE);
    fprintf (fout, "  switch (%s->rq_proc)\n  {\n", RQSTP);

    if (!nullproc(vp->procs))
    {
      fprintf (fout, "    case NULLPROC:\n");
      fprintf (fout, "         svc_sendreply (%s,xdr_void,NULL);\n", TRANSP);
      fprintf (fout, "         return;\n\n");
    }
    for (proc = vp->procs; proc; proc = proc->next)
    {
      fprintf (fout, "    case %s:\n", proc->proc_name);
      fprintf (fout, "         xdr_%s = xdr_%s;\n", ARG, stringfix(proc->arg_type));
      fprintf (fout, "         xdr_%s = xdr_%s;\n", RESULT, stringfix(proc->res_type));
      fprintf (fout, "         %s = (char *(*)()) ", ROUTINE);
      pvname (proc->proc_name, vp->vers_num);
      fprintf (fout, ";\n");
      fprintf (fout, "         break;\n\n");
    }
    fprintf (fout, "    default:\n");
    printerr ("noproc", TRANSP, 9);
    fprintf (fout, "         return;\n");
    fprintf (fout, "  }\n");

    fprintf (fout, "  memset ((void*)&%s,0,sizeof(%s));\n", ARG, ARG);
    printif ("getargs", TRANSP, "&", ARG);
    printerr ("decode", TRANSP, 4);
    fprintf (fout, "    return;\n  }\n");

    fprintf (fout, "  %s = (*%s)(&%s, %s);\n", RESULT, ROUTINE, ARG, RQSTP);
    fprintf (fout, "  if (%s && !svc_sendreply(%s, xdr_%s, %s))\n",
             RESULT, TRANSP, RESULT, RESULT);
    printerr ("systemerr", TRANSP, 5);
    fprintf (fout, "\n");

    printif ("freeargs", TRANSP, "&", ARG);
    fprintf (fout, "    fprintf (stderr, \"unable to free arguments\\n\");\n");
    fprintf (fout, "    return;\n  }\n}\n\n");
  }
}

void write_programs (char *storage)
{
  list       *l;
  definition *def;

  for (l = defined; l; l = l->next)
  {
    def = (definition *) l->val;
    if (def->def_kind == DEF_PROGRAM)
       write_program (def, storage);
  }
}

