/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#if !defined(lint) && defined(SCCSIDS)
static char sccsid[] = "@(#)clnt_perror.c 1.15 87/10/07 Copyr 1984 Sun Micro";
#endif

/*
 * clnt_perror.c
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <rpc/rpc.h>
#include <rpc/auth.h>
#include <rpc/clnt.h>

#include "rpc_mac.h"

#ifdef __HIGHC__
//#define sys_errlist _sys_errlist
//#define sys_nerr    _sys_nerr
#endif

struct auth_errtab {
       enum auth_stat status;
       char          *message;
     };

struct rpc_errtab {
       enum clnt_stat status;
       char          *message;
     };

static struct auth_errtab auth_errlist[] = {
  { AUTH_OK,              "Authentication OK"          },
  { AUTH_BADCRED,         "Invalid client credential"  },
  { AUTH_REJECTEDCRED,    "Server rejected credential" },
  { AUTH_BADVERF,         "Invalid client verifier"    },
  { AUTH_REJECTEDVERF,    "Server rejected verifier"   },
  { AUTH_TOOWEAK,         "Client credential too weak" },
  { AUTH_INVALIDRESP,     "Invalid server verifier"    },
  { AUTH_FAILED,          "Failed (unspecified error)" },
};

static struct rpc_errtab  rpc_errlist[] = {
  { RPC_SUCCESS,          "RPC: Success"                      },
  { RPC_CANTENCODEARGS,   "RPC: Can't encode arguments"       },
  { RPC_CANTDECODERES,    "RPC: Can't decode result"          },
  { RPC_CANTSEND,         "RPC: Unable to send"               },
  { RPC_CANTRECV,         "RPC: Unable to receive"            },
  { RPC_TIMEDOUT,         "RPC: Timed out"                    },
  { RPC_VERSMISMATCH,     "RPC: Incompatible versions of RPC" },
  { RPC_AUTHERROR,        "RPC: Authentication error"         },
  { RPC_PROGUNAVAIL,      "RPC: Program unavailable"          },
  { RPC_PROGVERSMISMATCH, "RPC: Program/version mismatch"     },
  { RPC_PROCUNAVAIL,      "RPC: Procedure unavailable"        },
  { RPC_CANTDECODEARGS,   "RPC: Server can't decode arguments"},
  { RPC_SYSTEMERROR,      "RPC: Remote system error"          },
  { RPC_UNKNOWNHOST,      "RPC: Unknown host"                 },
  { RPC_UNKNOWNPROTO,     "RPC: Unknown protocol"             },
  { RPC_PMAPFAILURE,      "RPC: Port mapper failure"          },
  { RPC_PROGNOTREGISTERED,"RPC: Program not registered"       },
  { RPC_FAILED,           "RPC: Failed (unspecified error)"   }
};


static char *buf = NULL;

static char *_buf (void)
{
  if (buf == NULL)
     buf = malloc (1024);   /* !! was 256 */
  return (buf);
}

static char *auth_errmsg (enum auth_stat stat)
{
  int i;
  for (i = 0; i < sizeof(auth_errlist)/sizeof(struct auth_errtab); i++)
  {
    if (auth_errlist[i].status == stat)
       return (auth_errlist[i].message);
  }
  return (NULL);
}

/*
 * Print reply error info
 */
char * clnt_sperror (CLIENT *rpch, char *s)
{
  struct rpc_err e;
  char  *err;
  char  *str      = _buf();
  char  *strstart = str;

  if (str == 0)
     return (NULL);

  CLNT_GETERR (rpch, &e);

  str += sprintf (str, "%s: %s", s, clnt_sperrno(e.re_status));

  switch (e.re_status)
  {
    case RPC_SUCCESS:
    case RPC_CANTENCODEARGS:
    case RPC_CANTDECODERES:
    case RPC_TIMEDOUT:
    case RPC_PROGUNAVAIL:
    case RPC_PROCUNAVAIL:
    case RPC_CANTDECODEARGS:
    case RPC_SYSTEMERROR:
    case RPC_UNKNOWNHOST:
    case RPC_UNKNOWNPROTO:
    case RPC_PMAPFAILURE:
    case RPC_PROGNOTREGISTERED:
    case RPC_FAILED:
         break;

    case RPC_CANTSEND:
    case RPC_CANTRECV:
         str += sprintf (str, "; errno = %s", sys_errlist[e.re_errno]);
         break;

    case RPC_VERSMISMATCH:
         str += sprintf (str, "; low version = %lu, high version = %lu",
                         e.re_vers.low, e.re_vers.high);
         break;

    case RPC_AUTHERROR:
         err = auth_errmsg (e.re_why);
         str += sprintf (str,"; why = ");
         if (err)
              str += sprintf (str, "%s",err);
         else str += sprintf (str, "(unknown authentication error - %d)", e.re_why);
         break;

    case RPC_PROGVERSMISMATCH:
         str += sprintf (str, "; low version = %lu, high version = %lu",
                         e.re_vers.low, e.re_vers.high);
         break;

    default:  /* unknown */
         str += sprintf (str, "; s1 = %lu, s2 = %lu",
                         e.re_lb.s1, e.re_lb.s2);
         break;
  }
  strcat (str, "\n");
  return (strstart) ;
}

void clnt_perror (CLIENT *rpch, char *s)
{
  (*_printf) ("%s",clnt_sperror(rpch,s));
}


/*
 * This interface for use by clntrpc
 */
char * clnt_sperrno (enum clnt_stat stat)
{
  int i;
  for (i = 0; i < DIM(rpc_errlist); i++)
  {
    if (rpc_errlist[i].status == stat)
       return (rpc_errlist[i].message);
  }
  return ("RPC: (unknown error code)");
}

void clnt_perrno (enum clnt_stat num)
{
  (*_printf) ("%s",clnt_sperrno(num));
}


char * clnt_spcreateerror (char *s)
{
  char *p, *str = _buf();

  if (!str)
     return (NULL);

  p = str;
  p += sprintf (p, "%s: %s", s, clnt_sperrno(rpc_createerr.cf_stat));

  switch (rpc_createerr.cf_stat)
  {
    case RPC_PMAPFAILURE:
         p += sprintf (p, " - %s", clnt_sperrno(rpc_createerr.cf_error.re_status));
         break;

    case RPC_SYSTEMERROR:
         if (rpc_createerr.cf_error.re_errno > 0 &&
             rpc_createerr.cf_error.re_errno < sys_nerr)
              p += sprintf (p, " - %s",
                            sys_errlist[rpc_createerr.cf_error.re_errno]);
         else p += sprintf (p, " - Error %d",
                            rpc_createerr.cf_error.re_errno);
         break;
    default:
         break;
  }
  strcat (str,"\n");
  return (str);
}

void clnt_pcreateerror (char *s)
{
  (*_printf) ("%s",clnt_spcreateerror(s));
}


