/*
 *  Some handy macros for the implementation of DOS-RPC
 */

#ifndef __RPC_MACROS_H
#define __RPC_MACROS_H

#include <assert.h>
#include <netinet/in.h>  /* for sockaddr_in */
#include <tcp.h>         /* Waterloo TCP main definitions */

#if defined(__DJGPP__) || defined(__HIGHC__) || defined(__WATCOMC__)
#include <unistd.h>
#endif

#if defined (__TURBOC__) || (defined (__WATCOMC__) && !defined(__386__))
#define __16BIT__
#endif

#if defined (__HIGHC__) || defined (__DJGPP__) || (defined (__WATCOMC__) && defined(__386__))
#define __32BIT__
#endif

#if !defined(__MSDOS__) && (defined(__32BIT__) || defined(__16BIT__))
#define __MSDOS__
#endif


#ifdef __DJGPP__
#define SLASH   '/'
#else
#define SLASH   '\\'
#endif

#define RPC_NOSUPPORT       255            /* error */
#define MAX_BROADCAST_SIZE  1400
#define LASTUNSIGNED        ((u_int)0-1)

#define ARGSUSED(x)    (x = x)
#define _DIAGASSERT(x) assert(x)
#define DIM(x)         sizeof(x)/sizeof(x[0])
#define Abort()        (*_printf)("RPC abort at %s line %d\r\n",\
                         __FILE__,__LINE__), exit(1)

#ifdef __HIGHC__
#  define MIN(a,b)     _min(a,b)
#  define MAX(a,b)     _max(a,b)
#else
#  define MIN(a,b)     ((a) < (b) ? (a) : (b))
#  define MAX(a,b)     ((a) > (b) ? (a) : (b))
#endif

#ifdef DEBUG
#  define MSG(str)     outsnl(str)
#  define MSG1(a,b)    outs(a),outsnl(b)
#else
#  define MSG(str)     ((void)0)
#  define MSG1(a,b)    ((void)0)
#endif

#ifndef IPPORT_TIMESERVER
#define IPPORT_TIMESERVER 37
#endif

#ifndef NGROUPS
#define NGROUPS 2
#endif

typedef struct rpc_config {
        char *key_messenger;
        char *rpc_programs;
        char *rpc_rootkey;
        char *rpc_keystore;
        int   pmap_local;
        int   rtime_port;
      } RPC_CONFIG;

extern RPC_CONFIG _rpc_cfg;

/*
 *  Prototypes
 */

extern int  getnetname    (char *name);
extern int  user2netname  (char *netname, int uid, char *domain);
extern int  host2netname  (char *netname, char *host, char *domain);

extern int  rtime         (struct sockaddr_in*, struct timeval*, struct timeval*);
extern void get_myaddress (struct sockaddr_in*);
extern int _rpc_dtablesize(void);
extern int  bindresvport  (int sd, struct sockaddr_in *sin);

extern void passwd2des    (char *pw, char *key);
extern int  xencrypt      (char *secret, char *passwd);
extern int  xdecrypt      (char *secret, char *passwd);


#ifdef __RPC_RPC_H
extern bool_t xdr_opaque_auth    (XDR *xdrs, struct opaque_auth *ap);
extern bool_t xdr_accepted_reply (XDR *xdrs, struct accepted_reply *ar);
extern bool_t xdr_rejected_reply (XDR *xdrs, struct rejected_reply *rr);
#endif

/*
 * WatTcp portmapper service
 */
extern int pmap_start_server (void);
extern int pmap_stop_server  (void);

#endif
